import sys
import os
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'src')))

import pytest
from unittest.mock import mock_open, patch
from pathlib import Path
import yaml
import yamale
from yamale.validators import Validator, DefaultValidators

from ryo.src.utils import (
    load_config,
    get_task,
    config_validator,
)

# --- Mock de Configurações ---
mock_valid_config_yaml = """
tasks:
  example-task:
    repository: minha_lib\\workflow
    steps:
      - action: commit_file
        auto_commit: 'true'
        commit_msg: "Segundo repositorio"
        repository: minha_lib\\teste3\\workflow

      - action: commit_file
        auto_commit: 'true'

      - action: workflow_monitor
        workflow_name: "1 - FEAT - Build and PR"
        show_workflow: 'true'

      - action: approve_pull_request
        source_branch: feature-teste
        target_branch: develop

      - action: workflow_monitor
        workflow_name: "2 - DEV - Build and Deploy"
        show_workflow: 'true'

      - action: approve_pull_request
        source_branch: develop
        target_branch: release/*

      - action: workflow_monitor
        workflow_name: "3 - HOM - Homologacao"
        show_workflow: 'true'

      - action: approve_pull_request
        source_branch: release/*
        target_branch: main
"""

mock_config_dict = yaml.safe_load(mock_valid_config_yaml)

mock_invalid_config_yaml = """
tasks:
  example-task:
    repository: minha_lib\\workflow
    steps:
      - action: unknown_action
        file_path: 'README.md'
        commit_msg: 'Update README.md'
"""

expected_load_config_data = mock_config_dict

# --- Testes para load_config ---
def test_load_config():
    with patch('ryo.src.utils.open', mock_open(read_data=mock_valid_config_yaml)):
        config = load_config()
        assert config == expected_load_config_data

# --- Testes para get_task ---
class TestGetTask:
    def test_existing_task(self):
        with patch('ryo.src.utils.load_config', return_value=mock_config_dict):
            task = get_task("example-task")
            assert task == mock_config_dict["tasks"]["example-task"]

    def test_non_existing_task(self):
        with patch('ryo.src.utils.load_config', return_value=mock_config_dict):
            with pytest.raises(KeyError, match="Task 'non_existent_task' not founded."):
                get_task("non_existent_task")

# --- Testes para config_validator ---
# @pytest.mark.parametrize(
#     "yaml_content, expected",
#     [
#         (
#             """
#             tasks:
#               example-task:
#                 repository: minha_lib\workflow
#                 steps:
#                   - action: commit_file
#                     auto_commit: 'true'
#                     commit_msg: "Segundo repositorio"
#                     repository: minha_lib\teste3\workflow

#                   - action: commit_file
#                     auto_commit: 'true'

#                   - action: workflow_monitor
#                     workflow_name: "1 - FEAT - Build and PR"
#                     show_workflow: 'true'

#                   - action: approve_pull_request
#                     source_branch: feature-teste
#                     target_branch: develop

#                   - action: workflow_monitor
#                     workflow_name: "2 - DEV - Build and Deploy"
#                     show_workflow: 'true'

#                   - action: approve_pull_request
#                     source_branch: develop
#                     target_branch: release/*

#                   - action: workflow_monitor
#                     workflow_name: "3 - HOM - Homologacao"
#                     show_workflow: 'true'

#                   - action: approve_pull_request
#                     source_branch: release/*
#                     target_branch: main
#             """,
#             True,
#         ),
#         # ... (outros casos de teste para config_validator) ...
#     ],
# )
# def test_config_validator(yaml_content, expected):
#     with patch("builtins.open", mock_open(read_data=yaml_content)):
#         with patch("os.path.join", return_value="dummy_path"):
#             # Mock da leitura do schema.yml
#             mock_schema_content = f"""
#                   tasks: map(include('task'), required=True)

#                   ---
#                   task:
#                     repository: str(required=True)
#                     steps: list(include('step'), required=True)

#                   ---
#                   step: include('step_definition')

#                   ---
#                   step_definition:
#                     action: str(required=True)
#                     file_path: str(required=False)
#                     commit_msg: str(required=False)
#                     workflow_name: str(required=False)
#                     show_workflow: str(required=False)
#                     source_branch: str(required=False)
#                     target_branch: str(required=False)
#             """
#             with patch("builtins.open", mock_open(read_data=mock_schema_content), create=True) as mock_schema_open:
#                 # Forçar o retorno do caminho correto para o schema
#                 with patch("os.path.dirname", return_value="dummy_schema_dir"):
#                     with patch("os.path.abspath", return_value="dummy_schema_dir/test_utils.py"):
#                         with patch("os.path.join", side_effect=lambda *args: "/".join(args) if "schema.yml" in args else "dummy_path"):
#                             result = config_validator("dummy_base_dir")
#                             assert result == expected


import os
import yaml
import yamale
from yamale.validators import Validator, DefaultValidators
from pathlib import Path


class StepValidator(Validator):
    tag = "step"

    def _is_valid(self, value):
        action_fields = {
            "commit_file": {
                "required": [],
                "optional": ["auto_commit", "commit_msg", "repository"],
            },
            "workflow_monitor": {
                "required": ["workflow_name"],
                "optional": ["show_workflow"],
            },
            "approve_pull_request": {
                "required": ["source_branch", "target_branch"],
                "optional": [],
            },
        }

        action = value.get("action")
        if not action:
            self._errors.append("O campo 'action' é obrigatório.")
            return False

        if action not in action_fields:
            self._errors.append(f"A ação '{action}' não é reconhecida.")
            return False

        required_fields = action_fields[action]["required"]
        optional_fields = action_fields[action]["optional"]
        allowed_fields = set(required_fields + optional_fields + ["action"])

        missing_fields = [field for field in required_fields if field not in value]
        if missing_fields:
            self._errors.append(
                f"Campos obrigatórios ausentes para a ação '{action}': {', '.join(missing_fields)}."
            )
            return False

        unexpected_fields = [field for field in value if field not in allowed_fields]
        if unexpected_fields:
            self._errors.append(
                f"Campos inesperados para a ação '{action}': {', '.join(unexpected_fields)}."
            )
            return False

        return True

    def fail(self, value):
        return " ".join(self._errors)


def config_validator(base_dir: str) -> bool:
    """
    Valida um arquivo YAML contra um esquema definido.

    Args:
        base_dir (str): Base dir.

    Returns:
        bool: True se o arquivo for válido, False caso contrário.
    """
    config_path = os.path.join(base_dir, ".config.yml")
    utils_base_dir = os.path.dirname(os.path.abspath(__file__))
    schema_path = os.path.join(utils_base_dir, "schema.yml")

    try:
        with open(config_path, "r", encoding="utf-8") as f:
            content = f.read()
        data = yamale.make_data(content=content)
        validators = DefaultValidators.copy()
        validators[StepValidator.tag] = StepValidator
        schema = yamale.make_schema(schema_path, validators=validators)
        yamale.validate(schema, data)
        return True
    except FileNotFoundError:
        print(f"Erro: Arquivo de configuração não encontrado em: {config_path}")
        return False
    except yamale.YamaleError as exc:
        print(".config.yml não é válido:")
        for result in exc.results:
            for error in result.errors:
                print(f"  - {error}")
        return False
    except Exception as e:
        print(f"Ocorreu um erro inesperado durante a validação: {e}")
        return False


def check_config_name():
    config_files = ["config.yml", "config.yaml"]
    config_file = None

    for file in config_files:
        if os.path.exists(os.path.join(os.getcwd(), file)):
            config_file = file
            break

    if config_file is None:
        raise FileNotFoundError(
            "Config file not founded (.config.yml ou .config.yaml)."
        )

    return config_file


def load_config(config_file: Path) -> dict:
    """
    Load the configuration from the YAML file.

    Returns:
        dict: The loaded configuration.
    """
    # print("Getting config file...")

    config_path = os.path.join(os.getcwd(), config_file)

    with open(config_path, "r", encoding="utf-8") as f:
        try:
            config = yaml.safe_load(f)
            # print(f"Load configuration with success!")
            return config
        except FileNotFoundError:
            print(
                f"Aviso: arquivo config.yml não encontrado em: {os.getcwd()}. Usando configurações padrão (se houver)."
            )
            return {}
        except yaml.YAMLError as e:
            print(f"Erro ao ler o arquivo config.yml: {e}")
            return {}


def get_task(task_name: str, config_file: Path) -> dict:
    """
    Load the configuration from the YAML file and return the task details.

    Args:
        task_name (str): The name of the task to retrieve.

    Returns:
        dict: The task details.
    """
    # print("Getting task parameters...")
    try:
        config = load_config(config_file)
        task = config["tasks"].get(task_name)
        if task != None:
            return task
        else:
            raise KeyError(f"Task '{task_name}' not founded.")
    except KeyError as exc:  # Verificar: esta é a mensagem de erro correta?
        print(f"Error getting desired task: {exc}")
        raise



from ryo.github import GitHubRepo
from pathlib import Path
from ryo.aws import AWS


def commit_file(step: dict, repo: GitHubRepo):
    """
    Commit a file with a message.

    Args:
        step (dict): The step configuration containing file path and commit message.
    """
    file_path = step.get("file_path", "README.md")
    commit_msg = step.get("commit_msg", "Update README.md")
    auto_commit = step.get("auto_commit", "false")
    branch = step.get("branch", "")

    repo.commit(
        commit_file=file_path,
        commit_message=commit_msg,
        auto_commit=auto_commit,
        branch=branch,
    )


def approve_pull_request(step: dict, repo: GitHubRepo):
    """
    Approve a pull request.

    Args:
        step (dict): The step configuration containing target branch.
    """
    target_branch = step.get("target_branch", "develop")
    source_branch = step.get("source_branch", "")
    repo.approve_pull_request(target_branch=target_branch, source_branch=source_branch)


def workflow_monitor(step: dict, repo: GitHubRepo):
    """
    Monitor a workflow.

    Args:
        step (dict): The step configuration containing workflow name.
    """
    workflow_name = step.get("workflow_name")
    show_workflow = step.get("show_workflow", "false").lower()

    repo.workflow_monitor(workflow_name=workflow_name, show_workflow=show_workflow)


def git_clone(step: dict, repo: GitHubRepo):
    """
    Git Clone

    Args:
        step (dict): The step configuration containing workflow name.
    """
    branch = step.get("branch")

    repo.git_clone( branch=branch)


def replace_file(step: dict, repo: GitHubRepo):
    """
    Realiza o replace de um arquivo

    Args:
        step (dict): The step configuration containing workflow name.
    """
    source_path = Path(step.get("source_path"))
    target_path = Path(step.get("target_path"))

    repo.replace_file(source_path=source_path, target_path=target_path)


def replace_dir(step: dict, repo: GitHubRepo):
    """
    Realiza o replace de uma pasta

    Args:
        step (dict): The step configuration containing workflow name.
    """
    source_path = Path(step.get("source_path"))
    target_path = Path(step.get("target_path", ""))

    repo.replace_dir(source_path=source_path, target_path=target_path)


def destroy(step: dict, repo: GitHubRepo):
    """
    Realiza as alterações para que a esteira seja executada no modo de destroy

    Args:
        step (dict): The step configuration containing workflow name.
    """
    file_path = Path(step.get("file_path"))
    value = step.get("value")

    repo.destroy(file_path=file_path, value=value)


def start_step_function(step: dict):
    """
    Realiza o start da execução da step function

    Args:
        step (dict): The step configuration containing workflow name.
    """
    step_function_name = step.get("name")
    aws_account = step.get("account")
    aws_region = step.get("region", "sa-east-1")
    aws = AWS(
        aws_account=aws_account,
        step_function_name=step_function_name,
        aws_region=aws_region,
    )

    aws.start_step_function()



import boto3
from tenacity import retry
from tenacity.stop import stop_after_attempt
from tenacity.wait import wait_exponential, wait_fixed
from tenacity.retry import retry_if_result


@staticmethod
def retry_func(status):
    print("Current status:", status, flush=True)
    if status == "vfailed":
        raise Exception("Execution failed.")
    elif status == "timed_out":
        raise Exception("Execution timed out.")
    elif status == "aborted":
        raise Exception("Execution was aborted.")
    return status == "running"


class AWS:
    def __init__(self, aws_account, step_function_name, aws_region):
        self.aws_region = aws_region
        self.step_function_name = step_function_name
        self.aws_account = aws_account
        self.state_machine_arn = f"arn:aws:states:{aws_region}:{aws_account}:stateMachine:{step_function_name}"

    def get_client(self):
        try:
            print("Getting client ...")
            stepfunction_client = boto3.client(
                "stepfunctions", region_name=self.aws_region
            )
            return stepfunction_client
        except Exception as e:
            print("Error getting client:", e)

    def start_execution(self, client):
        try:
            print("Start execution ...")
            response = client.start_execution(stateMachineArn=self.state_machine_arn)
            print("Execution ARN:", response["executionArn"])
            return response["executionArn"]
        except Exception as e:
            print("Error starting execution:", e)

    @staticmethod
    def retry_func(status):
        print("Current status:", status, flush=True)
        if status == "vfailed":
            raise Exception("Execution failed.")
        elif status == "timed_out":
            raise Exception("Execution timed out.")
        elif status == "aborted":
            raise Exception("Execution was aborted.")
        return status == "running"

    def get_execution(self, execution_arn, client):
        print("ENTRANDO NO GET EXECUTION")
        try:
            response = client.describe_execution(executionArn=execution_arn)
            print(f"Status: {response['status'].lower()}")
            return response["status"].lower()
        except Exception as e:
            print("Error getting execution:", e)

    @retry(
        retry=retry_if_result(retry_func),
        wait=wait_fixed(10),
        stop=stop_after_attempt(3600),
    )
    def check_execution_status(self, client, execution_arn):

        status = self.get_execution(execution_arn=execution_arn, client=client)
        print(f"Status get execution: {status}")
        return status

    def start_step_function(self):
        """Realiza a alteracao de um arquivo dentro do repositorio."""
        print(f"\n------------- Iniciando o start da Step Function ------------")

        client = self.get_client()
        execution_arn = self.start_execution(client=client)
        final_status = self.check_execution_status(
            execution_arn=execution_arn, client=client
        )
        print("Final status:", final_status, flush=True)



from ryo.core import commit_file, workflow_monitor, approve_pull_request, git_clone
from ryo.core import replace_dir, replace_file, destroy, start_step_function
from pathlib import Path
from ryo.github import GitHubRepo


def run_steps(task: dict, base_path: str):
    """
    Execute the steps in the task.

    Args:
        task (dict): The task configuration.
    """
    task_repo = task.get("repository")
    task_org = task.get("org")
    steps = task.get("steps", [])

    for step in steps:
        action = step.get("action")
        step_repo = step.get("repository")
        repository = step_repo if step_repo else task_repo
        repo = GitHubRepo(repo_path=repository, base_path=base_path, org=task_org)
        if action == "commit_file":
            commit_file(step, repo)
        elif action == "workflow_monitor":
            workflow_monitor(step, repo)
        elif action == "approve_pull_request":
            approve_pull_request(step, repo)
        elif action == "clone":
            git_clone(step, repo)
        elif action == "replace_file":
            replace_file(step, repo)
        elif action == "replace_dir":
            replace_dir(step, repo)
        elif action == "destroy":
            destroy(step, repo)
        elif action == "start_step_function":
            start_step_function(step)
        else:
            print(f"Unknown action: {action}")




import os
import pytest
from unittest.mock import patch, mock_open, MagicMock
from ryo.src.core import GitHubRepo  # ajuste se o caminho do seu módulo for diferente


@pytest.fixture
def repo(tmp_path):
    base_path = tmp_path
    repo_path = tmp_path / "my-repo"
    repo_path.mkdir()
    return GitHubRepo(repo_path="my-repo", base_path=str(base_path))


def test_cd_repo_path_success(repo):
    assert repo.cd_repo_path() is True
    assert os.getcwd() == repo.repo_path


def test_cd_repo_path_fail():
    repo = GitHubRepo(repo_path="nonexistent", base_path="/invalid/path")
    assert repo.cd_repo_path() is None


@patch("subprocess.run")
def test_run_command_success(mock_run, repo):
    mock_run.return_value = MagicMock(stdout="success", stderr="")
    stdout, stderr = repo.run_command("echo 'Hello'")
    assert stdout == "success"
    assert stderr == ""


@patch("ryo.src.core.GitHubRepo.cd_repo_path", return_value=True)
@patch("ryo.src.core.GitHubRepo.run_command")
def test_commit_with_auto_commit(mock_run_command, mock_cd, repo, tmp_path):
    commit_file = tmp_path / "README.md"
    commit_file.write_text("Initial content")
    mock_run_command.side_effect = [
        ("", ""),  # git pull
        ("", ""),  # git status
        ("", ""),  # git add
        ("", ""),  # git commit
        ("", "")   # git push
    ]
    repo.commit(str(commit_file), "Auto commit test", "true")
    with open(commit_file) as f:
        assert f.read().endswith(" ")


@patch("ryo.src.core.GitHubRepo.cd_repo_path", return_value=True)
@patch("ryo.src.core.GitHubRepo.run_command")
def test_commit_no_auto_commit_with_no_changes(mock_run_command, mock_cd, repo, capsys):
    mock_run_command.side_effect = [
        ("", ""),  # git pull
        ("", "")   # git status returns empty, no changes
    ]
    repo.commit("dummy_file", "Message", "false")
    captured = capsys.readouterr()
    assert "Nenhuma alteração foi encontrada" in captured.out



import pytest
from unittest.mock import patch, MagicMock, call
import os
import subprocess
from ryo.src.github import GitHubRepo

@pytest.fixture
def github_repo():
    mock_base_path = "/tmp"
    mock_repo_path = "my_repo"
    with patch("os.path.abspath") as mock_abspath:
        mock_abspath.return_value = "/tmp/my_repo"
        repo = GitHubRepo(mock_repo_path, mock_base_path)
    return repo

class TestGitHubRepo:
    def test_init(self):
        mock_base_path = "/tmp"
        mock_repo_path = "my_repo"
        with patch("os.path.abspath") as mock_abspath:
            mock_abspath.return_value = "/tmp/my_repo"
            repo = GitHubRepo(mock_repo_path, mock_base_path)
            mock_abspath.assert_called_once_with(os.path.join(mock_base_path, mock_repo_path))
        assert repo.repo_path == "/tmp/my_repo"
        assert repo.repo_name == "my_repo"
        assert repo.base_path == "/tmp"

    def test_cd_repo_path_success(self, github_repo):
        with patch("os.chdir") as mock_chdir:
            result = github_repo.cd_repo_path()
            mock_chdir.assert_called_once_with("/tmp/my_repo")
            assert result is True

    def test_cd_repo_path_failure(self, github_repo):
        with patch("os.chdir", side_effect=FileNotFoundError) as mock_chdir:
            result = github_repo.cd_repo_path()
            mock_chdir.assert_called_once_with("/tmp/my_repo")
            assert result is None

    def test_run_command_success(self, github_repo):
        with patch("subprocess.run") as mock_run:
            mock_run.return_value = MagicMock(stdout="saida", stderr="", returncode=0)
            output, error = github_repo.run_command("echo test")
            mock_run.assert_called_once_with("echo test", shell=True, capture_output=True, text=True)
            assert output == "saida"
            assert error == ""

    def test_run_command_failure(self, github_repo):
        with patch("subprocess.run", side_effect=subprocess.CalledProcessError(1, "comando falho", stderr="erro")) as mock_run:
            output = github_repo.run_command("comando falho")
            mock_run.assert_called_once_with("comando falho", shell=True, capture_output=True, text=True)
            assert output is None

    def test_commit(self, github_repo):
        with patch("os.chdir") as mock_chdir, patch("subprocess.run") as mock_run, patch("builtins.open", create=True) as mock_open:
            mock_run.return_value = MagicMock(stdout="", stderr="", returncode=0)
            github_repo.commit("arquivo.txt", "mensagem de commit", "true")
            assert mock_chdir.call_count == 2
            mock_chdir.assert_any_call("/tmp")
            mock_chdir.assert_any_call("/tmp/my_repo")
            assert mock_run.call_count == 5
            mock_open.assert_called_once_with("arquivo.txt", "a")

    

    # def test_commit_no_auto_commit(self):
    #     github_repo = GitHubRepo(repo_path="my_repo", base_path="/tmp")

    #     with patch("os.chdir") as mock_chdir, patch.object(GitHubRepo, "run_command") as mock_run:
    #         mock_run.side_effect = [
    #             ("", ""),  # git pull
    #             ("", ""),  # git status --porcelain → sem alterações
    #         ]

    #         github_repo.commit("README.md", "mensagem de teste", "false")

    #         # Verificações
    #         assert mock_chdir.call_count == 1
    #         mock_chdir.assert_any_call("/tmp/my_repo")
    #         mock_chdir.assert_any_call("/tmp")

    #         assert mock_run.call_count == 1
    #         mock_run.assert_has_calls([
    #             call("git pull"),
    #             call("git status --porcelain"),
    #         ])


    def test_approve_pull_request_with_source_and_target(self, github_repo):
        with patch("os.chdir") as mock_chdir, patch("subprocess.run") as mock_run:
            mock_run.side_effect = [
                MagicMock(stdout="123", stderr="", returncode=0),  # gh pr list
                MagicMock(stdout="", stderr="", returncode=0)      # gh pr review
            ]
            github_repo.approve_pull_request("feature/branch", "main")
            mock_chdir.assert_called_once_with("/tmp/my_repo")
            assert mock_run.call_count == 2

    def test_approve_pull_request_with_only_target(self, github_repo):
        with patch("os.chdir") as mock_chdir, patch("subprocess.run") as mock_run:
            mock_run.return_value = MagicMock(stdout="develop", stderr="", returncode=0)
            github_repo.approve_pull_request("", "main")
            mock_chdir.assert_called_once_with("/tmp/my_repo")
            mock_run.assert_called()

    
    def test_approve_pull_request_target_branch_wildcard(self, github_repo):
        with patch("os.chdir") as mock_chdir, patch("subprocess.run") as mock_run:
            mock_run.side_effect = [
                MagicMock(stdout="123\n456", stderr="", returncode=0),  # gh pr list
                MagicMock(stdout="", stderr="", returncode=0)           # gh pr review
            ]
            github_repo.approve_pull_request("feature/xyz", "release/*")
            mock_chdir.assert_called_once_with("/tmp/my_repo")
            assert mock_run.call_count == 2

    def test_approve_pull_request_source_branch_wildcard(self, github_repo):
        with patch("os.chdir") as mock_chdir, patch("subprocess.run") as mock_run:
            mock_run.side_effect = [
                MagicMock(stdout="789\n1011", stderr="", returncode=0),
                MagicMock(stdout="", stderr="", returncode=0)
            ]
            github_repo.approve_pull_request("feature/*", "main")
            mock_chdir.assert_called_once_with("/tmp/my_repo")
            assert mock_run.call_count == 2

    def test_approve_pull_request_invalid_wildcard(self, github_repo):
        with patch("os.chdir") as mock_chdir, patch("subprocess.run") as mock_run:
            github_repo.approve_pull_request("feature/*", "release/*")
            mock_chdir.assert_called_once_with("/tmp/my_repo")
            mock_run.assert_not_called()

    def test_workflow_monitor(self, github_repo):
        with patch("os.chdir") as mock_chdir, patch("subprocess.run") as mock_run, patch("time.sleep") as mock_sleep:
            mock_run.side_effect = [
                MagicMock(stdout="in_progress\tresult\t\t\t\t\t12345", stderr="", returncode=0),
                MagicMock(stdout="completed\tsuccess\t\t\t\t\t12345", stderr="", returncode=0),
                MagicMock(stdout="", stderr="", returncode=0),
            ]
            github_repo.workflow_monitor("workflow", "true")
            mock_chdir.assert_called_once_with("/tmp/my_repo")
            assert mock_sleep.call_count == 1

    def test_workflow_monitor_completed_no_show(self, github_repo):
        with patch("os.chdir") as mock_chdir, patch("subprocess.run") as mock_run, patch("time.sleep") as mock_sleep:
            mock_run.return_value = MagicMock(stdout="completed\tsuccess\t\t\t\t\t512345", stderr="", returncode=0)
            github_repo.workflow_monitor("workflow", "false")
            mock_chdir.assert_called_once_with("/tmp/my_repo")
            mock_sleep.assert_not_called()

    def test_workflow_monitor_workflow_not_found(self, github_repo):
        with patch("os.chdir") as mock_chdir, patch("subprocess.run") as mock_run:
            mock_run.return_value = MagicMock(stdout="", stderr="", returncode=0)
            github_repo.workflow_monitor("workflow-inexistente", "true")
            mock_chdir.assert_called_once_with("/tmp/my_repo")
            mock_run.assert_called_once()



import os
import subprocess
import time
from pathlib import Path
import shutil
import yaml
import sys
import os
from datetime import datetime
from typing import List
import re

class SSOAuthorizationRequired(Exception):
    def __init__(self, url, message="SSO authorization required."):
        self.url = url
        self.message = message
        super().__init__(f"{message} URL: {url}")


class GitHubRepo:
    def __init__(self, repo_path, base_path, org):
        self.repo_path = os.path.abspath(os.path.join(base_path, repo_path))
        self.repo_name = repo_path.split("/")[-1]
        self.base_path = base_path
        self.org = org

    def cd_repo_path(self):
        """
        Changes the current working directory to the repository path.

        Returns:
            bool: True if the directory change is successful.

        Raises:
            FileNotFoundError: If the repository path does not exist.
            OSError: For other OS-related errors.
        """
        try:
            os.chdir(self.repo_path)
            return True
        except FileNotFoundError:
            raise FileNotFoundError(f"Repository path not found: {self.repo_path}")
        except OSError as e:
            raise OSError(f"Failed to change directory to '{self.repo_path}': {e}")


    def run_command(self, command):
        """Executa um comando no shell e retorna a saída."""
        try:
            result = subprocess.run(
                command, shell=True, capture_output=True, text=True, check=True
            )
            return result.stdout.strip(), result.stderr.strip()
        except subprocess.CalledProcessError as e:
            match = re.search(r"https://github\.com/enterprises/.+?/sso\?authorization_request=[^\s]+", e.stderr)
            if match:
                raise SSOAuthorizationRequired(url=match.group(0))
            else:
                print(f"Erro ao executar comando: {command}\n")
                print("STDOUT:", e.stdout)
                print("STDERR:", e.stderr)
                raise


    def checkout_or_create_branch(self, branch: str) -> None:
        """Faz checkout de uma branch existente ou cria nova a partir da 'develop'."""
        if not branch:
            branch = datetime.now().strftime("%Y%m%d%H%M%S")

        try:
            out, err = self.run_command(f"git branch --list {branch}")
            branch_exists = out.strip() or err.strip()

            if branch_exists:
                print(f"📍 Branch '{branch}' encontrada localmente. Fazendo checkout...")
                self.run_command(f"git checkout {branch}")
            else:
                print(f"\n🆕 Branch '{branch}' não existe. Criando a partir da 'develop'...")
                self.run_command(f"git checkout -b {branch} develop")

        except Exception as e:
            print(f"❌ Erro ao verificar ou criar a branch '{branch}': {e}")
            raise


    def get_current_branch(self):
        try:
            source_branch, _ = self.run_command("git rev-parse --abbrev-ref HEAD")
            return source_branch
        except Exception as e:
            print(f"Erro durante a busca da branch: {e}")
            raise

    def git_clone(self, branch: str) -> None:
        """Clona o repositório e troca/cria a branch especificada."""
        try:
            print(f"\n------------- Realizando o clone do repositorio -------------")
            
            if self.repo_exists():
                print("------------- Repositório já existe localmente -------------")
            else:
                self.clone_repo()
                self.cd_repo_path()
                self.checkout_or_create_remote_branch(branch)

            print("--------------- Clone realizado com sucesso--- -------------")

        except Exception as e:
            print(f"❌ Erro durante o processo de clone: {e}")
            raise
        finally:
            os.chdir(self.base_path)

    def repo_exists(self) -> bool:
        """Verifica se o repositório já foi clonado."""
        return Path(self.repo_name).is_dir()

    def clone_repo(self) -> None:
        """Clona o repositório remoto."""
        repo_url = f"https://github.com/{self.org}/{self.repo_name}.git"
        out, err = self.run_command(f"git clone {repo_url}")
        

    def checkout_or_create_remote_branch(self, branch: str) -> None:
        """Verifica e faz checkout da branch remota ou cria uma nova."""
        out, err = self.run_command("git branch -r")
        remote_branches = out + err

        if f"origin/{branch}" in remote_branches:
            print(f"Branch remota '{branch}' encontrada. Fazendo checkout...")
            self.run_command(f"git checkout -b {branch} origin/{branch}")
        else:
            print(f"Branch '{branch}' não existe remotamente. Criando nova...")
            self.run_command(f"git checkout -b {branch}")
            self.run_command(f"git push -u origin {branch}")
            print("✅ Branch criada e enviada com sucesso")


    def commit(self, commit_file, commit_message, auto_commit, branch):
        """Realiza um commit no arquivo README.md para execução dos workflows."""
        try:
            print(f"\n Processando repositório: {self.repo_name}")

            print(f"\n------------------------ Run Commit -------------------------")

            self.cd_repo_path()

            print(f"------------------- Verificando a branch --------------------")
            self.checkout_or_create_branch(branch)
            branch = self.get_current_branch()

            print("------------------- Realizando o git pull -------------------")

            self.run_command("git pull")

            status_output, _ = self.run_command("git status --porcelain")
            if status_output:
                print("Existem alterações pendentes no repositório.")
            elif auto_commit.lower() == "true":
                print("-------------------- Alterando o arquivo --------------------")
                with open(commit_file, "a") as file:
                    file.write(" ")
            else:
                print("-------------- Nenhuma alteração foi encontrada -------------")
                return

            print("------------- Adicionando o arquivo ao staging --------------")
            self.run_command("git add .")

            print("-------- Realizando o commit com a mensagem fornecida -------")
            self.run_command(f'git commit -m "{commit_message}"')

            print("--------------------- Realizando o push ---------------------")
            self.run_command("git push -u")
            print("---------------- Commit realizado com sucesso ---------------")
        except Exception as e:
            print(f"Erro durante o commit: {e}")
            raise
        finally:
            os.chdir(self.base_path)

    def _maybe_show_workflow(self, workflow_id: str, show_workflow: str):
        if show_workflow.lower() == "true":
            print(f"Exibindo detalhes do workflow: {workflow_id}")
            self.run_command(
                f"gh run view {workflow_id} -w --repo {self.org}/{self.repo_name}"
            )

    def workflow_monitor(self, workflow_name, show_workflow):
        """Monitora a execução de um workflow."""
        try:
            print(f"\n------------------- Run Workflow Monitor --------------------")

            print(f"Monitorando o workflow: {workflow_name}")
            time.sleep(30)

            while True:
                last_workflow_execution, _ = self.run_command(
                    f"gh run list --repo {self.org}/{self.repo_name} | grep '{workflow_name}' | head -n 1 "
                )

                if last_workflow_execution:
                    elementos = last_workflow_execution.split("\t")
                    status = elementos[0].strip()
                    workflow_id = elementos[6].strip()

                    if status == "completed":
                        print(
                            "--------------- Workflow executado com sucesso -------------"
                        )
                        if show_workflow.lower() == "true":
                            print(f"Exibindo detalhes do workflow: {workflow_id}")
                            self._maybe_show_workflow(workflow_id, show_workflow)
                        break
                    elif status == "failure":
                        print(
                            "---------------- Workflow executado com falha --------------"
                        )
                        self._maybe_show_workflow(workflow_id, show_workflow)
                        raise RuntimeError("Workflow execution failed.")
                    elif status == "cancelled" or status == "skipped":
                        print(f"O workflow foi {status}.")
                        raise RuntimeError(f"Workflow was {status}.")
                    elif (
                        status == "pending"
                        or status == "queued"
                        or status == "in_progress"
                        or status == "waiting"
                    ):
                        print(
                            f"O workflow está {status}. Aguardando 30 segundos antes de verificar novamente."
                        )
                        time.sleep(30)
                    else:
                        print(f"Status desconhecido: {status}")
                        break
                else:
                    print(f"Erro: workflow '{workflow_name}' não encontrado.")
                    break

        except Exception as e:
            print(f"Erro durante a execução do workflow: {e}")
            raise

    
    def replace_file(self, source_path, target_path):
        """Realiza a alteracao de um arquivo dentro do repositorio."""
        try:
            print(f"\n-------------- Iniciando a alteracao do arquivo -------------")

            target_path = Path(self.base_path) / Path(self.repo_path) / target_path
            source_path_clean = str(source_path).lstrip("/\\")
            source_path_fix = Path(self.base_path) / source_path_clean

            # print(f"-------------- source path : {source_path_fix} -------------")
            # print(f"-------------- target path : {target_path} -------------")

            if source_path_fix.exists():
                # Cria o diretório destino se não existir
                target_path.parent.mkdir(parents=True, exist_ok=True)

                shutil.copyfile(source_path_fix, target_path)
                print(f"------------- O arquivo foi alterado com sucesso ------------")
            else:
                print(f"------------- O arquivo de origem não existe -----------")
                raise FileNotFoundError(
                    f"Source: {source_path_fix}, Target: {target_path}"
                )

        except Exception as e:
            print(f"Erro durante a alteração do arquivo: {e}")
            raise

    def replace_dir(self, source_path, target_path):
        """Realiza a alteracao de um arquivo dentro do repositorio."""
        try:
            print(f"\n------------- Iniciando a alteracao dos arquivos ------------")
            source_path = str(source_path).lstrip("/\\")
            source_path = Path(self.base_path) / source_path
            target_path = str(target_path).lstrip("/\\")
            target_path = Path(self.base_path) / Path(self.repo_path) / target_path

            # print(f"-------------- source path : {source_path} -------------")
            # print(f"-------------- target path : {target_path} -------------")

            if source_path.is_dir() and target_path.is_dir():
                shutil.copytree(source_path, target_path, dirs_exist_ok=True)
                print(
                    f"------------ Os arquivos foram alterados com sucesso ----------"
                )
            else:
                print(
                    f"--------------- Uma ou ambas as pastas não existem ------------"
                )
                raise
        except Exception as e:
            print(f"Erro durante a alteração dos arquivos: {e}")
            raise

    def destroy(self, file_path, value):
        """Realiza a alteracao do parametro destroy para deletar os recursos."""
        print(f"\n-------- Iniciando a alteracao do parametro destroy ---------")
        try:
            self.cd_repo_path()
            with open(file_path, "r") as f:
                config = yaml.safe_load(f)

            config["infra"]["terraform"]["destroy"] = str(value)

            with open(file_path, "w") as f:
                yaml.dump(config, f, default_flow_style=False)
                print(f"-------------- Alteração realizada com sucesso --------------")

        except Exception as e:
            print(f" Erro inesperado: {e} ")
            raise
        finally:
            os.chdir(self.base_path)

    def get_all_workflow_executions(self, workflow: str) -> List:
        try:
            response, response2 = self.run_command(f'gh run list --workflow "{workflow}" --repo {self.org}/{self.repo_name} ')
            return response
        except Exception as e:
            print(f"Erro ao obter o status do workflow: {e}")
            raise

    
    def get_job_id(self, workflow_id: str) -> List:
        try:
            response, response2 = self.run_command(f'gh run view "{workflow_id}" --repo {self.org}/{self.repo_name} ')
            return response
        except Exception as e:
            print(f"Erro ao obter o job ID do workflow: {e}")
            raise


    def rerun_workflow(self, workflow_id: str):
        try:
            print(f"\n----------------- Reexecutando o workflow -------------------")
            response, response2 = self.run_command(f'gh run rerun "{workflow_id}" --repo {self.org}/{self.repo_name} --failed')
            print(f"-------------- Workflow reexecutado com sucesso --------------")
            return True
        except Exception as e:
            print(f"Erro ao reexecutar o workflow: {e}")
            raise

    def approve_pull_request(self, source_branch: str, target_branch: str, repository=None):
        try:
            if repository == None:
                repository = self.repo_name

            print(f"\n----------------- Run Approve Pull Request ------------------")
            print(f"Verificando a existência de PR '{source_branch}' -> '{target_branch }'")        

            if source_branch == "":
                self.cd_repo_path()
                source_branch, _ = self.run_command("git rev-parse --abbrev-ref HEAD")
                print("Branch atual:", source_branch)
                os.chdir(self.base_path)

            pr_number = self.get_pr_id(source_branch, target_branch, repository)
            
            if pr_number:
                print(f"------------- Um PR foi encontrado: #{pr_number} ------------")
                self.run_command(f"gh pr merge {pr_number} --merge --repo {self.org}/{repository}")
                print("----------------- Merge realizado com sucesso ---------------")
            else:
                print("------------ Nenhum PR encontrado foi encontrado ------------")
                return
        except Exception as e:
            print(f"Erro durante a aprovação do pr: {e}")
            raise

    def get_pr_id(self, source_branch: str, target_branch: str, repository=None):
        try:
            if repository == None:
                repository = self.repo_name

            pr_number = ""
            if "*" in source_branch and "*" in target_branch:
                print("Pelo menos um dos parâmetros não pode conter '*'")
                return

            if "*" not in source_branch and "*" not in target_branch:
                command = [
                    'gh', 'pr', 'list',
                    '--head', source_branch,
                    '--base', target_branch,
                    '--repo', f'{self.org}/{repository}',
                    '--limit', '1',
                    '--json', 'number',
                    '--jq', '.[0].number'
                ]
                pr_number, _ = self.run_command(command)

            elif "*" in target_branch:
                target_branch_base = target_branch.rstrip("*")
                command = [
                    'gh', 'pr', 'list',
                    '--head', source_branch,
                    '--repo', f'{self.org}/{repository}',
                    '--json', 'number,baseRefName',
                    '--jq', f'.[] | select(.baseRefName | startswith("{target_branch_base}")) | .number'
                ]
                pr_number = self.__parse_branch_wildcard(command, 'baseRefName')

            elif "*" in source_branch:
                source_branch_base = source_branch.rstrip("*")
                command = [
                    'gh', 'pr', 'list',
                    '--base', target_branch,
                    '--repo', f'{self.org}/{repository}',
                    '--json', 'number,headRefName',
                    '--jq', f'.[] | select(.headRefName | startswith("{source_branch_base}")) | .number'
                ]
                pr_number = self.__parse_branch_wildcard(command, 'headRefName')
            return pr_number
        except Exception as e:
            print(f"Erro durante a aprovação do pr: {e}")
            raise        


    def __parse_branch_wildcard(self, command):
        pr_numbers, _ = self.run_command(command)
        if pr_numbers:
            numeros_str = pr_numbers.strip().split('\n')
            numeros_int = [int(num) for num in numeros_str if num.strip()]
            return max(numeros_int)
        return ""
    


import pytest
from unittest.mock import patch, MagicMock
from pathlib import Path
from ryo.src.runner import run_steps
from ryo.src.github import GitHubRepo


@pytest.fixture
def base_path(tmp_path):
    return str(tmp_path)


@pytest.fixture
def task():
    return {
        "repository": "my-repo",
        "steps": [
            {"action": "commit_file", "repository": "my-repo", "commit_message": "Test commit", "auto_commit": "true", "commit_file": "README.md"},
            {"action": "workflow_monitor", "repository": "my-repo", "workflow_name": "CI", "show_workflow": "false"},
            {"action": "approve_pull_request", "repository": "my-repo", "source_branch": "feature-branch", "target_branch": "main"}
        ]
    }


@patch("ryo.src.core.commit_file")
@patch("ryo.src.core.workflow_monitor")
@patch("ryo.src.core.approve_pull_request")
@patch("ryo.src.github.GitHubRepo")
def test_run_steps(mock_GitHubRepo, mock_approve, mock_workflow, mock_commit, task, base_path):
    # Mock GitHubRepo behavior
    mock_repo_instance = MagicMock()
    mock_GitHubRepo.return_value = mock_repo_instance

    # Chamando run_steps para simular a execução
    run_steps(task, base_path)

    # Verificar se as funções foram chamadas
    mock_commit.assert_called_once()
    mock_workflow.assert_called_once()
    mock_approve.assert_called_once()

    # Verificar se as funções receberam os parâmetros corretos
    commit_call_args = mock_commit.call_args[0][0]
    assert commit_call_args.get("commit_message") == "Test commit"
    assert commit_call_args.get("commit_file") == "README.md"

    workflow_call_args = mock_workflow.call_args[0][0]
    assert workflow_call_args.get("workflow_name") == "CI"

    approve_call_args = mock_approve.call_args[0][0]
    assert approve_call_args.get("source_branch") == "feature-branch"
    assert approve_call_args.get("target_branch") == "main"


@patch("ryo.src.core.commit_file")
@patch("ryo.src.core.workflow_monitor")
@patch("ryo.src.core.approve_pull_request")
@patch("ryo.src.github.GitHubRepo")
def test_run_steps_unknown_action(mock_GitHubRepo, mock_approve, mock_workflow, mock_commit, task, base_path):
    # Adiciona uma etapa com ação desconhecida
    task["steps"].append({"action": "unknown_action", "repository": "my-repo"})

    with patch("builtins.print") as mock_print:
        run_steps(task, base_path)
        mock_print.assert_called_with("Unknown action: unknown_action")


@patch("ryo.src.core.commit_file")
@patch("ryo.src.core.workflow_monitor")
@patch("ryo.src.core.approve_pull_request")
@patch("ryo.src.github.GitHubRepo")
def test_run_steps_missing_action(mock_GitHubRepo, mock_approve, mock_workflow, mock_commit, task, base_path):
    # Remove a chave "action" de uma das etapas
    del task["steps"][0]["action"]

    with patch("builtins.print") as mock_print:
        run_steps(task, base_path)
        mock_print.assert_called_with("Unknown action: None")



import yaml
import os
import argparse
from ryo.utils import get_task
from ryo.runner import run_steps

__version__ = "0.1.7"


def main():
    parser = argparse.ArgumentParser(
        description="Um utilitário de linha de comando Ryo."
    )
    subparsers = parser.add_subparsers(dest="comando", help="Subcomandos disponíveis")

    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {__version__}"
    )

    task_parser = subparsers.add_parser("task", help="Acompanha a tarefa")
    task_parser.add_argument("nome", help="O nome da task a ser implementada")
    task_parser.add_argument(
        "--file",
        help="Caminho do arquivo de configuração YAML (default: config.yml)",
        default="config.yml",
    )

    args = parser.parse_args()

    if args.comando == "task":
        task = get_task(args.nome, config_file=args.file)
        if task:
            run_steps(task, os.getcwd())
        else:
            print(f"Task '{args.nome}' não encontrada.")

    else:
        parser.print_help()


if __name__ == "__main__":
    main()



from setuptools import setup, find_packages

setup(
    name='ryo',
    version='0.1.7',
    packages=find_packages(),
    install_requires=['PyYAML', 'yamale', 'boto3', 'tenacity'],
    entry_points={
        'console_scripts': [
            'ryo=ryo.cli:main', 
        ],
    },
    author='Ana Victoria de Oliveira Silva',
    author_email='anavictoriaavos@gmail.com',
    description='A personal CLI tool using YAML and schema validation.',
    license='MIT',
    python_requires='>=3.7',
)
